/* File: $Id: blockdev.e,v 1.19 1998/02/25 16:17:56 morsiani Exp $ */

/****************************************************************************
 *
 * External declarations from the corresponding module .cc
 *
 ****************************************************************************/

// This class implements the block devices' 512 byte sectors/tape blocks.
// Each object contains a single buffer; methods are provided to read/write
// these blocks from/to real files and to access to the word-sized contents.
// This class is provided primarily to make DMA transfer easier and to 
// standardize block handling. 

class Block
{
	public:
		
		// This method returns an empty (unitialized) 512 byte Block
		Block(void);
		
		// Object deletion is done by default handler
		
		// This method fills a Block with file contents starting at "offset"
		// bytes from file start, as computed by caller.
		// Returns TRUE if read does not succeed, FALSE otherwise
		Boolean ReadBlock(FILE * blkFile, SWord offset);

		// This method writes Block contents in a file, starting at "offset"
		// bytes from file start, as computed by caller. Returns TRUE if
		// write does not succeed, FALSE otherwise
		Boolean WriteBlock(FILE * blkFile, SWord offset);
		
		// This method returns the Word contained in the Block at ofs (Word
		// items) offset, range [0..BLOCKSIZE - 1]. Warning: in-bounds
		// checking is leaved to caller
		Word getWord(unsigned int ofs);

		// This method fills with "value" the Word contained in the Block at
		// ofs (Word items) offset, range [0..BLOCKSIZE - 1]. Warning:
		// in-bounds checking is leaved to caller
		void setWord(unsigned int ofs, Word value);
			
	private:
		// Block contents
		Word blkBuf[BLOCKSIZE];
};


/****************************************************************************/


// This class contains the simulated disk drive geometry and performance
// parameters. They are filled by mkdev utility and used by DiskDevice class
// for detailed disk performance simulation.
// Position, min, max and default values, where applicable, are defined in 
// h/blockdev.h header file.
//
// Parameters are:
// number of cylinders;
// number of heads;
// number of sectors per track;
// disk rotation time in microseconds;
// average track-to-track seek time in microseconds;
// data % of sector (to compute inter-sector gap)

class DriveParams
{
	public:
		
		// This method reads from disk parameters from file header, builds a
		// DriveParams object, and returns the disk sectors start offset:
		// this allows to modify the parameters' size without changing the
		// caller.  If fileOfs returned is 0, something has gone wrong; file
		// is rewound after use
		DriveParams(FILE * diskFile, SWord * fileOfs);
		
		// Object deletion is done by default handler		

		// These methods return the corresponding geometry or performance 
		// figure 
		unsigned int getCylNum(void);
		unsigned int getHeadNum(void);
		unsigned int getSectNum(void);
		unsigned int getRotTime(void);
		unsigned int getSeekTime(void);
		unsigned int getDataSect(void);
		
		private:
			// parameter buffer
			unsigned int parms[DRIVEPNUM];
};
