/* File: $Id: bus.support.e,v 1.19 1998/02/25 16:17:56 morsiani Exp $ */

/****************************************************************************
 *
 * External declarations from the corresponding module .cc
 *
 ****************************************************************************/

class TimeStamp;


// Event class is used to keep track of the external events of the
// system: device operations and interrupt generation.
// Every object contains a device number saying what device will complete
// its operation and generate a interrupt, a TimeStamp saying when it 
// will happen, and a pointer to the next event in a structure, if needed
// (set to NULL otherwise) 

class Event
{
	public: 
		
		// This method creates a new Event object and initalizes it
		Event(TimeStamp * ts, Word inc, unsigned int il, unsigned int dev);
		
		// This method deletes a Event object and its internal structures
		~Event(void);
		
		// This method returns the interrupt line of the device requiring
		// the Event
		unsigned int getIntLine(void);
		
		// This method returns the device number of the device requiring
		// the Event
		unsigned int getDevNum(void);

		// This method returns the TimeStamp access pointer
		TimeStamp * getTS(void);
		
		// This method links an Event to its successor in a structure
		void AddBefore(Event * ev);
		
		// This method inserts an Event after another, linking the former to the
		// successor of the latter
		void InsAfter(Event * ev);
		
		// This method returns the pointer to the successor of an Event
		Event * Next(void);
	
	private:

		// interrupt line and device number identifying the device which 
		// requires the Event
		unsigned int intL;
		unsigned int devNum;

		// Event verification time
		TimeStamp * time;

		Event * next;
};


/****************************************************************************/


// This class implements a NULL-terminated sorted queue of Event objects,
// used to schedule the device events in the system.
// An object contains a pointer to the head of the queue, and a pointer to
// the last inserted Event (to speed up insertion).
// The queue is sorted on ascending timestamp order

class EventQueue
{
	public:
		
		// This method creates a new (empty) queue
		EventQueue(void);
		
		// This method deletes the queue and its associated structures
		~EventQueue(void);
		
		// This method returns TRUE if the queue is empty, FALSE otherwise
		Boolean IsEmptyQ(void);
		
		// This method returns the TimeStamp in the EventQueue
		// head, if queue is not empty, and NULL otherwise
		TimeStamp * getHTS(void);

		// This method returns the interrupt line of the EventQueue head, if
		// not empty (0 otherwise)
		unsigned int getHIntLine(void);
		
		// This method returns the device number of the EventQueue head, if
		// not empty (0 otherwise)
		unsigned int getHDevNum(void);

		// This method creates a new Event object and inserts it in the
		// EventQueue; EventQueue is sorted on ascending time order
		TimeStamp * InsertQ(TimeStamp * ts, Word inc, unsigned int il, unsigned int dev);
		
		// This method removes the head of a (not empty) queue and sets it to the
		// following Event		
		void RemoveHead(void);
		
	private:
		
		// head of the queue
		Event * head;
		
		// last Event inserted
		Event * lastIns;
};
